@@ -35,9 +35,13 @@
 #include <event.h>
 #include <fcntl.h>
 #include <errno.h>
+#include <imsg.h>
+#include <sha1.h>
+#include <sha2.h>
 
 #include "got_opentemp.h"
 #include "got_reference.h"
+#include "got_object.h"
 
 #include "gotwebd.h"
 #include "log.h"
@@ -49,8 +53,7 @@ config_init(struct gotwebd *env)
 
 	strlcpy(env->httpd_chroot, D_HTTPD_CHROOT, sizeof(env->httpd_chroot));
 
-	env->prefork_gotwebd = GOTWEBD_NUMPROC;
-	env->server_cnt = 0;
+	env->prefork = GOTWEBD_NUMPROC;
 	TAILQ_INIT(&env->servers);
 	TAILQ_INIT(&env->sockets);
 	TAILQ_INIT(&env->addresses);
@@ -68,20 +71,11 @@ int
 config_getcfg(struct gotwebd *env, struct imsg *imsg)
 {
 	/* nothing to do but tell gotwebd configuration is done */
-	if (sockets_compose_main(env, IMSG_CFG_DONE, NULL, 0) == -1)
+	if (sockets_compose_main(env, GOTWEBD_IMSG_CFG_DONE, NULL, 0) == -1)
 		fatal("sockets_compose_main IMSG_CFG_DONE");
 	return 0;
 }
 
-int
-config_setserver(struct gotwebd *env, struct server *srv)
-{
-	if (main_compose_sockets(env, IMSG_CFG_SRV, -1, srv, sizeof(*srv))
-	    == -1)
-		fatal("main_compose_sockets IMSG_CFG_SRV");
-	return 0;
-}
-
 int
 config_getserver(struct gotwebd *env, struct imsg *imsg)
 {
@@ -106,17 +100,20 @@ config_getserver(struct gotwebd *env, struct imsg *imsg)
 }
 
 int
-config_setsock(struct gotwebd *env, struct socket *sock)
+config_setsock(struct gotwebd *env, struct socket *sock, uid_t uid, gid_t gid)
 {
 	/* open listening sockets */
-	if (sockets_privinit(env, sock) == -1)
+	if (sockets_privinit(env, sock, uid, gid) == -1)
 		return -1;
 
-	if (main_compose_sockets(env, IMSG_CFG_SOCK, sock->fd,
+	if (main_compose_sockets(env, GOTWEBD_IMSG_CFG_SOCK, sock->fd,
 	    &sock->conf, sizeof(sock->conf)) == -1)
 		fatal("main_compose_sockets IMSG_CFG_SOCK");
 
-	sock->fd = -1;
+	if (main_compose_gotweb(env, GOTWEBD_IMSG_CFG_SOCK, sock->fd,
+	    &sock->conf, sizeof(sock->conf)) == -1)
+		fatal("main_compose_gotweb GOTWEBD_IMSG_CFG_SOCK");
+
 	return 0;
 }
 
@@ -133,7 +130,7 @@ config_getsock(struct gotwebd *env, struct imsg *imsg)
 	memcpy(&sock_conf, p, sizeof(sock_conf));
 
 	if (IMSG_DATA_SIZE(imsg) != sizeof(sock_conf)) {
-		log_debug("%s: imsg size error", __func__);
+		log_warnx("%s: imsg size error", __func__);
 		return 1;
 	}
 
@@ -164,21 +161,21 @@ config_setfd(struct gotwebd *env)
 {
 	int i, j, fd;
 
-	log_debug("%s: Allocating %d file descriptors",
+	log_info("%s: Allocating %d file descriptors",
 	    __func__, PRIV_FDS__MAX + GOTWEB_PACK_NUM_TEMPFILES);
 
 	for (i = 0; i < PRIV_FDS__MAX + GOTWEB_PACK_NUM_TEMPFILES; i++) {
-		for (j = 0; j < env->nserver; ++j) {
+		for (j = 0; j < env->prefork; j++) {
 			fd = got_opentempfd();
 			if (fd == -1)
 				fatal("got_opentemp");
-			if (imsg_compose_event(&env->iev_server[j],
-			    IMSG_CFG_FD, 0, -1, fd, NULL, 0) == -1)
-				fatal("imsg_compose_event IMSG_CFG_FD");
+			if (imsg_compose_event(&env->iev_gotweb[j],
+			    GOTWEBD_IMSG_CFG_FD, 0, -1, fd, NULL, 0) == -1)
+				fatal("imsg_compose_event GOTWEBD_IMSG_CFG_FD");
 
-			if (imsgbuf_flush(&env->iev_server[j].ibuf) == -1)
-				fatal("imsg_flush");
-			imsg_event_add(&env->iev_server[j]);
+			if (imsgbuf_flush(&env->iev_gotweb[j].ibuf) == -1)
+				fatal("imsgbuf_flush");
+			imsg_event_add(&env->iev_gotweb[j]);
 		}
 	}
 
