@@ -1,7 +1,6 @@
 /*
  * Copyright (c) 2016, 2019, 2020-2022 Tracey Emery <tracey@traceyemery.net>
  * Copyright (c) 2015 Mike Larkin <mlarkin@openbsd.org>
- * Copyright (c) 2014 Reyk Floeter <reyk@openbsd.org>
  * Copyright (c) 2013 David Gwynne <dlg@openbsd.org>
  * Copyright (c) 2013 Florian Obser <florian@openbsd.org>
  *
@@ -25,7 +24,6 @@
 #include <sys/stat.h>
 #include <sys/types.h>
 
-#include <ctype.h>
 #include <dirent.h>
 #include <errno.h>
 #include <event.h>
@@ -54,38 +52,6 @@
 #include "log.h"
 #include "tmpl.h"
 
-static const struct querystring_keys querystring_keys[] = {
-	{ "action",		ACTION },
-	{ "commit",		COMMIT },
-	{ "file",		RFILE },
-	{ "folder",		FOLDER },
-	{ "headref",		HEADREF },
-	{ "index_page",		INDEX_PAGE },
-	{ "path",		PATH },
-};
-
-static const struct action_keys action_keys[] = {
-	{ "blame",	BLAME },
-	{ "blob",	BLOB },
-	{ "blobraw",	BLOBRAW },
-	{ "briefs",	BRIEFS },
-	{ "commits",	COMMITS },
-	{ "diff",	DIFF },
-	{ "error",	ERR },
-	{ "index",	INDEX },
-	{ "patch",	PATCH },
-	{ "summary",	SUMMARY },
-	{ "tag",	TAG },
-	{ "tags",	TAGS },
-	{ "tree",	TREE },
-	{ "rss",	RSS },
-};
-
-static const struct got_error *gotweb_init_querystring(struct querystring **);
-static const struct got_error *gotweb_parse_querystring(struct querystring *,
-    char *);
-static const struct got_error *gotweb_assign_querystring(struct querystring *,
-    char *, char *);
 static int gotweb_render_index(struct template *);
 static const struct got_error *gotweb_load_got_path(struct repo_dir **,
     const char *, struct request *);
@@ -96,7 +62,6 @@ static const struct got_error *gotweb_get_repo_description(char **,
 static const struct got_error *gotweb_get_clone_url(char **, struct server *,
     const char *, int);
 
-static void gotweb_free_querystring(struct querystring *);
 static void gotweb_free_repo_dir(struct repo_dir *);
 
 struct server *gotweb_get_server(const char *);
@@ -154,6 +119,18 @@ gotweb_get_socket(int sock_id)
 	return NULL;
 }
 
+static void
+cleanup_request(struct request *c)
+{
+	uint32_t request_id = c->request_id;
+
+	fcgi_cleanup_request(c);
+
+	if (imsg_compose_event(gotwebd_env->iev_sockets, GOTWEBD_IMSG_REQ_ABORT,
+	    GOTWEBD_PROC_GOTWEB, -1, -1, &request_id, sizeof(request_id)) == -1)
+		log_warn("imsg_compose_event");
+}
+
 static struct request *
 recv_request(struct imsg *imsg)
 {
@@ -209,14 +186,14 @@ recv_request(struct imsg *imsg)
 	c->tp = template(c, fcgi_write, c->outbuf, GOTWEBD_CACHESIZE);
 	if (c->tp == NULL) {
 		log_warn("gotweb init template");
-		fcgi_cleanup_request(c);
+		cleanup_request(c);
 		return NULL;
 	}
 
 	c->sock = gotweb_get_socket(c->sock_id);
 	if (c->sock == NULL) {
 		log_warn("socket id '%d' not found", c->sock_id);
-		fcgi_cleanup_request(c);
+		cleanup_request(c);
 		return NULL;
 	}
 
@@ -224,15 +201,15 @@ recv_request(struct imsg *imsg)
 	error = gotweb_init_transport(&c->t);
 	if (error) {
 		log_warnx("gotweb init transport: %s", error->msg);
-		fcgi_cleanup_request(c);
+		cleanup_request(c);
 		return NULL;
 	}
 
 	/* get the gotwebd server */
-	srv = gotweb_get_server(c->server_name);
+	srv = gotweb_get_server(c->fcgi_params.server_name);
 	if (srv == NULL) {
-		log_warnx("server '%s' not found", c->server_name);
-		fcgi_cleanup_request(c);
+		log_warnx("server '%s' not found", c->fcgi_params.server_name);
+		cleanup_request(c);
 		return NULL;
 	}
 	c->srv = srv;
@@ -253,48 +230,43 @@ gotweb_process_request(struct request *c)
 	size_t len;
 	int r, binary = 0;
 
-	/* parse our querystring */
-	error = gotweb_init_querystring(&qs);
-	if (error) {
-		log_warnx("%s: %s", __func__, error->msg);
-		goto err;
-	}
+	/* querystring */
+	qs = &c->fcgi_params.qs;
 	c->t->qs = qs;
-	error = gotweb_parse_querystring(qs, c->querystring);
-	if (error) {
-		log_warnx("%s: %s", __func__, error->msg);
-		goto err;
-	}
 
 	/* Log the request. */
 	if (gotwebd_env->gotwebd_verbose > 0) {
+		struct gotwebd_fcgi_params *p = &c->fcgi_params;
 		char *server_name = NULL;
-		char *querystring = NULL;
 		char *document_uri = NULL;
+		const char *action_name = NULL;
 
-		if (c->server_name[0] &&
-		    stravis(&server_name, c->server_name, VIS_SAFE) == -1) {
+		if (p->server_name[0] &&
+		    stravis(&server_name, p->server_name, VIS_SAFE) == -1) {
 			log_warn("stravis");
 			server_name = NULL;
 		}
-		if (c->querystring[0] &&
-		    stravis(&querystring, c->querystring, VIS_SAFE) == -1) {
-			log_warn("stravis");
-			querystring = NULL;
-		}
-		if (c->document_uri[0] &&
-		    stravis(&document_uri, c->document_uri, VIS_SAFE) == -1) {
+
+		if (p->document_uri[0] &&
+		    stravis(&document_uri, p->document_uri, VIS_SAFE) == -1) {
 			log_warn("stravis");
 			document_uri = NULL;
 		}
 
-		log_info("processing request: server='%s' query='%s' "
-		    "document_uri='%s'",
+		action_name = gotweb_action_name(qs->action);
+		log_info("processing request: server='%s' action='%s' "
+		    "commit='%s', file='%s', folder='%s', headref='%s' "
+		    "index_page=%d path='%s' document_uri='%s'",
 		    server_name ? server_name : "",
-		    querystring ? querystring : "",
+		    action_name ? action_name : "",
+		    qs->commit,
+		    qs->file,
+		    qs->folder,
+		    qs->headref,
+		    qs->index_page,
+		    qs->path,
 		    document_uri ? document_uri : "");
 		free(server_name);
-		free(querystring);
 		free(document_uri);
 	}
 
@@ -307,14 +279,14 @@ gotweb_process_request(struct request *c)
 	if (qs->action == BLAME || qs->action == BLOB ||
 	    qs->action == BLOBRAW || qs->action == DIFF ||
 	    qs->action == PATCH) {
-		if (qs->commit == NULL) {
+		if (qs->commit[0] == '\0') {
 			error = got_error(GOT_ERR_BAD_QUERYSTRING);
 			goto err;
 		}
 	}
 
 	if (qs->action != INDEX) {
-		if (qs->path == NULL) {
+		if (qs->path[0] == '\0') {
 			error = got_error(GOT_ERR_BAD_QUERYSTRING);
 			goto err;
 		}
@@ -326,7 +298,7 @@ gotweb_process_request(struct request *c)
 	}
 
 	if (qs->action == BLOBRAW || qs->action == BLOB) {
-		if (qs->folder == NULL || qs->file == NULL) {
+		if (qs->folder[0] == '\0' || qs->file[0] == '\0') {
 			error = got_error(GOT_ERR_BAD_QUERYSTRING);
 			goto err;
 		}
@@ -343,7 +315,7 @@ gotweb_process_request(struct request *c)
 
 	switch (qs->action) {
 	case BLAME:
-		if (qs->folder == NULL || qs->file == NULL) {
+		if (qs->folder[0] == '\0' || qs->file[0] == '\0') {
 			error = got_error(GOT_ERR_BAD_QUERYSTRING);
 			goto err;
 		}
@@ -360,10 +332,10 @@ gotweb_process_request(struct request *c)
 			struct gotweb_url url = {
 				.index_page = -1,
 				.action = BLOBRAW,
-				.path = qs->path,
-				.commit = qs->commit,
-				.folder = qs->folder,
-				.file = qs->file,
+				.path = qs->path[0] ? qs->path : NULL,
+				.commit = qs->commit[0] ? qs->commit : NULL,
+				.folder = qs->folder[0] ? qs->folder : NULL,
+				.file = qs->file[0] ? qs->file : NULL,
 			};
 
 			return gotweb_reply(c, 302, NULL, &url);
@@ -478,11 +450,12 @@ gotweb_process_request(struct request *c)
 		}
 		qs->action = SUMMARY;
 		commit = TAILQ_FIRST(&c->t->repo_commits);
-		if (commit && qs->commit == NULL) {
-			qs->commit = strdup(commit->commit_id);
-			if (qs->commit == NULL) {
-				error = got_error_from_errno("strdup");
-				log_warn("%s: strdup", __func__);
+		if (commit && qs->commit[0] == '\0') {
+			if (strlcpy(qs->commit, commit->commit_id,
+			    sizeof(qs->commit)) >= sizeof(qs->commit)) {
+				error = got_error_msg(GOT_ERR_NO_SPACE,
+				    "commit ID too long");
+				log_warn("%s: %s", __func__, error->msg);
 				goto err;
 			}
 		}
@@ -566,211 +539,6 @@ gotweb_init_transport(struct transport **t)
 	return error;
 }
 
-static const struct got_error *
-gotweb_init_querystring(struct querystring **qs)
-{
-	const struct got_error *error = NULL;
-
-	*qs = calloc(1, sizeof(**qs));
-	if (*qs == NULL)
-		return got_error_from_errno2(__func__, "calloc");
-
-	(*qs)->headref = strdup("HEAD");
-	if ((*qs)->headref == NULL) {
-		free(*qs);
-		*qs = NULL;
-		return got_error_from_errno2(__func__, "strdup");
-	}
-
-	(*qs)->action = INDEX;
-
-	return error;
-}
-
-static const struct got_error *
-gotweb_parse_querystring(struct querystring *qs, char *qst)
-{
-	const struct got_error *error = NULL;
-	char *tok1 = NULL, *tok1_pair = NULL, *tok1_end = NULL;
-	char *tok2 = NULL, *tok2_pair = NULL, *tok2_end = NULL;
-
-	if (qst == NULL)
-		return error;
-
-	tok1 = strdup(qst);
-	if (tok1 == NULL)
-		return got_error_from_errno2(__func__, "strdup");
-
-	tok1_pair = tok1;
-	tok1_end = tok1;
-
-	while (tok1_pair != NULL) {
-		strsep(&tok1_end, "&");
-
-		tok2 = strdup(tok1_pair);
-		if (tok2 == NULL) {
-			free(tok1);
-			return got_error_from_errno2(__func__, "strdup");
-		}
-
-		tok2_pair = tok2;
-		tok2_end = tok2;
-
-		while (tok2_pair != NULL) {
-			strsep(&tok2_end, "=");
-			if (tok2_end) {
-				error = gotweb_assign_querystring(qs, tok2_pair,
-				    tok2_end);
-				if (error)
-					goto err;
-			}
-			tok2_pair = tok2_end;
-		}
-		free(tok2);
-		tok1_pair = tok1_end;
-	}
-	free(tok1);
-	return error;
-err:
-	free(tok2);
-	free(tok1);
-	return error;
-}
-
-/*
- * Adapted from usr.sbin/httpd/httpd.c url_decode.
- */
-static const struct got_error *
-gotweb_urldecode(char *url)
-{
-	char		*p, *q;
-	char		 hex[3];
-	unsigned long	 x;
-
-	hex[2] = '\0';
-	p = q = url;
-
-	while (*p != '\0') {
-		switch (*p) {
-		case '%':
-			/* Encoding character is followed by two hex chars */
-			if (!isxdigit((unsigned char)p[1]) ||
-			    !isxdigit((unsigned char)p[2]) ||
-			    (p[1] == '0' && p[2] == '0'))
-				return got_error(GOT_ERR_BAD_QUERYSTRING);
-
-			hex[0] = p[1];
-			hex[1] = p[2];
-
-			/*
-			 * We don't have to validate "hex" because it is
-			 * guaranteed to include two hex chars followed by nul.
-			 */
-			x = strtoul(hex, NULL, 16);
-			*q = (char)x;
-			p += 2;
-			break;
-		default:
-			*q = *p;
-			break;
-		}
-		p++;
-		q++;
-	}
-	*q = '\0';
-
-	return NULL;
-}
-
-static const struct got_error *
-gotweb_assign_querystring(struct querystring *qs, char *key, char *value)
-{
-	const struct got_error *error = NULL;
-	const char *errstr;
-	int a_cnt, el_cnt;
-
-	error = gotweb_urldecode(value);
-	if (error)
-		return error;
-
-	for (el_cnt = 0; el_cnt < nitems(querystring_keys); el_cnt++) {
-		if (strcmp(key, querystring_keys[el_cnt].name) != 0)
-			continue;
-
-		switch (querystring_keys[el_cnt].element) {
-		case ACTION:
-			for (a_cnt = 0; a_cnt < nitems(action_keys); a_cnt++) {
-				if (strcmp(value, action_keys[a_cnt].name) != 0)
-					continue;
-				qs->action = action_keys[a_cnt].action;
-				break;
-			}
-			if (a_cnt == nitems(action_keys))
-				qs->action = ERR;
-			break;
-		case COMMIT:
-			qs->commit = strdup(value);
-			if (qs->commit == NULL) {
-				error = got_error_from_errno2(__func__,
-				    "strdup");
-				goto done;
-			}
-			break;
-		case RFILE:
-			qs->file = strdup(value);
-			if (qs->file == NULL) {
-				error = got_error_from_errno2(__func__,
-				    "strdup");
-				goto done;
-			}
-			break;
-		case FOLDER:
-			qs->folder = strdup(value);
-			if (qs->folder == NULL) {
-				error = got_error_from_errno2(__func__,
-				    "strdup");
-				goto done;
-			}
-			break;
-		case HEADREF:
-			free(qs->headref);
-			qs->headref = strdup(value);
-			if (qs->headref == NULL) {
-				error = got_error_from_errno2(__func__,
-				    "strdup");
-				goto done;
-			}
-			break;
-		case INDEX_PAGE:
-			if (*value == '\0')
-				break;
-			qs->index_page = strtonum(value, INT64_MIN,
-			    INT64_MAX, &errstr);
-			if (errstr) {
-				error = got_error_from_errno3(__func__,
-				    "strtonum", errstr);
-				goto done;
-			}
-			if (qs->index_page < 0)
-				qs->index_page = 0;
-			break;
-		case PATH:
-			qs->path = strdup(value);
-			if (qs->path == NULL) {
-				error = got_error_from_errno2(__func__,
-				    "strdup");
-				goto done;
-			}
-			break;
-		}
-
-		/* entry found */
-		break;
-	}
-done:
-	return error;
-}
-
 void
 gotweb_free_repo_tag(struct repo_tag *rt)
 {
@@ -800,19 +568,6 @@ gotweb_free_repo_commit(struct repo_commit *rc)
 	free(rc);
 }
 
-static void
-gotweb_free_querystring(struct querystring *qs)
-{
-	if (qs != NULL) {
-		free(qs->commit);
-		free(qs->file);
-		free(qs->folder);
-		free(qs->headref);
-		free(qs->path);
-	}
-	free(qs);
-}
-
 static void
 gotweb_free_repo_dir(struct repo_dir *repo_dir)
 {
@@ -844,7 +599,7 @@ gotweb_free_transport(struct transport *t)
 		gotweb_free_repo_tag(rt);
 	}
 	gotweb_free_repo_dir(t->repo_dir);
-	gotweb_free_querystring(t->qs);
+	t->qs = NULL;
 	free(t->more_id);
 	free(t->tags_more_id);
 	if (t->blob)
@@ -872,7 +627,7 @@ gotweb_index_navs(struct request *c, struct gotweb_url *prev, int *have_prev,
     struct gotweb_url *next, int *have_next)
 {
 	struct transport *t = c->t;
-	struct querystring *qs = t->qs;
+	const struct querystring *qs = t->qs;
 	struct server *srv = c->srv;
 
 	*have_prev = *have_next = 0;
@@ -902,7 +657,7 @@ gotweb_render_index(struct template *tp)
 	struct request *c = tp->tp_arg;
 	struct server *srv = c->srv;
 	struct transport *t = c->t;
-	struct querystring *qs = t->qs;
+	const struct querystring *qs = t->qs;
 	struct repo_dir *repo_dir = NULL;
 	struct dirent **sd_dent = t->repos;
 	unsigned int d_i, d_disp = 0;
@@ -1047,6 +802,8 @@ const char *
 gotweb_action_name(int action)
 {
 	switch (action) {
+	case NO_ACTION:
+		return "no action";
 	case BLAME:
 		return "blame";
 	case BLOB:
@@ -1112,10 +869,14 @@ gotweb_render_url(struct request *c, struct gotweb_url *url)
 	}
 
 	if (url->folder) {
-		tmp = gotweb_urlencode(url->folder);
-		if (tmp == NULL)
-			return -1;
-		r = tp_writef(c->tp, "%sfolder=%s", sep, tmp);
+		if (got_path_is_root_dir(url->folder))
+			tmp = NULL;
+		else {
+			tmp = gotweb_urlencode(url->folder);
+			if (tmp == NULL)
+				return -1;
+		}
+		r = tp_writef(c->tp, "%sfolder=%s", sep, tmp ? tmp : "");
 		free(tmp);
 		if (r == -1)
 			return -1;
@@ -1158,17 +919,45 @@ int
 gotweb_render_absolute_url(struct request *c, struct gotweb_url *url)
 {
 	struct template	*tp = c->tp;
-	const char	*proto = c->https ? "https" : "http";
+	struct gotwebd_fcgi_params *p = &c->fcgi_params;
+	const char	*proto = p->https ? "https" : "http";
 
 	if (tp_writes(tp, proto) == -1 ||
 	    tp_writes(tp, "://") == -1 ||
-	    tp_htmlescape(tp, c->server_name) == -1 ||
-	    tp_htmlescape(tp, c->document_uri) == -1)
+	    tp_htmlescape(tp, p->server_name) == -1 ||
+	    tp_htmlescape(tp, p->document_uri) == -1)
 		return -1;
 
 	return gotweb_render_url(c, url);
 }
 
+/* 
+ * Ensure that a path sent in the request will not escape from the given
+ * parent directory. This matters for got-portable where we are not
+ * necessarily running in chroot and cannot be protected by unveil(2).
+ */
+static const struct got_error *
+validate_path(const char *path, const char *parent_path,
+    const char *orig_path)
+{
+	const struct got_error *error = NULL;
+	char *abspath;
+
+	abspath = realpath(path, NULL);
+	if (abspath == NULL) {
+		/* Paths which cannot be resolved are safe. */
+		if (errno == ENOENT || errno == EACCES || errno == ENOTDIR)
+			return NULL;
+		return got_error_from_errno("realpath");
+	}
+
+	if (!got_path_is_child(abspath, parent_path, strlen(parent_path)))
+		error = got_error_path(orig_path, GOT_ERR_NOT_GIT_REPO);
+
+	free(abspath);
+	return error;
+}
+
 static const struct got_error *
 gotweb_load_got_path(struct repo_dir **rp, const char *dir,
     struct request *c)
@@ -1189,6 +978,12 @@ gotweb_load_got_path(struct repo_dir **rp, const char *dir,
 	    GOTWEB_GIT_DIR) == -1)
 		return got_error_from_errno("asprintf");
 
+	error = validate_path(dir_test, srv->repos_path, dir);
+	if (error) {
+		free(dir_test);
+		return error;
+	}
+
 	dt = opendir(dir_test);
 	if (dt == NULL) {
 		free(dir_test);
@@ -1375,10 +1170,9 @@ gotweb_shutdown(void)
 {
 	imsgbuf_clear(&gotwebd_env->iev_parent->ibuf);
 	free(gotwebd_env->iev_parent);
-	if (gotwebd_env->iev_server) {
-		imsgbuf_clear(&gotwebd_env->iev_server->ibuf);
-		free(gotwebd_env->iev_server);
-	}
+
+	imsgbuf_clear(&gotwebd_env->iev_sockets->ibuf);
+	free(gotwebd_env->iev_sockets);
 
 	while (!TAILQ_EMPTY(&gotwebd_env->servers)) {
 		struct server *srv;
@@ -1388,6 +1182,14 @@ gotweb_shutdown(void)
 		free(srv);
 	}
 
+	while (!TAILQ_EMPTY(&gotwebd_env->sockets)) {
+		struct socket *sock;
+
+		sock = TAILQ_FIRST(&gotwebd_env->sockets);
+		TAILQ_REMOVE(&gotwebd_env->sockets, sock, entry);
+		free(sock);
+	}
+
 	free(gotwebd_env);
 
 	exit(0);
@@ -1423,8 +1225,8 @@ gotweb_launch(struct gotwebd *env)
 	struct server *srv;
 	const struct got_error *error;
 
-	if (env->iev_server == NULL)
-		fatal("server process not connected");
+	if (env->iev_sockets == NULL)
+		fatal("sockets process not connected");
 
 #ifndef PROFILE
 	if (pledge("stdio rpath recvfd sendfd proc exec unveil", NULL) == -1)
@@ -1443,7 +1245,7 @@ gotweb_launch(struct gotwebd *env)
 	if (unveil(NULL, NULL) == -1)
 		fatal("unveil");
 
-	event_add(&env->iev_server->ev, NULL);
+	event_add(&env->iev_sockets->ev, NULL);
 }
 
 static void
@@ -1491,7 +1293,7 @@ gotweb_dispatch_server(int fd, short event, void *arg)
 				}
 
 				fcgi_create_end_record(c);
-				fcgi_cleanup_request(c);
+				cleanup_request(c);
 			}
 			break;
 		default:
@@ -1517,11 +1319,6 @@ recv_server_pipe(struct gotwebd *env, struct imsg *imsg)
 	struct imsgev *iev;
 	int fd;
 
-	if (env->iev_server != NULL) {
-		log_warn("server pipe already received");
-		return;
-	}
-
 	fd = imsg_get_fd(imsg);
 	if (fd == -1)
 		fatalx("invalid server pipe fd");
@@ -1529,7 +1326,6 @@ recv_server_pipe(struct gotwebd *env, struct imsg *imsg)
 	iev = calloc(1, sizeof(*iev));
 	if (iev == NULL)
 		fatal("calloc");
-
 	if (imsgbuf_init(&iev->ibuf, fd) == -1)
 		fatal("imsgbuf_init");
 	imsgbuf_allow_fdpass(&iev->ibuf);
@@ -1539,7 +1335,7 @@ recv_server_pipe(struct gotwebd *env, struct imsg *imsg)
 	event_set(&iev->ev, fd, EV_READ, gotweb_dispatch_server, iev);
 	imsg_event_add(iev);
 
-	env->iev_server = iev;
+	env->iev_sockets = iev;
 }
 
 static void
@@ -1617,8 +1413,9 @@ gotweb(struct gotwebd *env, int fd)
 
 	sockets_rlimit(-1);
 
-	if ((env->iev_parent = malloc(sizeof(*env->iev_parent))) == NULL)
-		fatal("malloc");
+	env->iev_parent = calloc(1, sizeof(*env->iev_parent));
+	if (env->iev_parent == NULL)
+		fatal("calloc");
 	if (imsgbuf_init(&env->iev_parent->ibuf, fd) == -1)
 		fatal("imsgbuf_init");
 	imsgbuf_allow_fdpass(&env->iev_parent->ibuf);
