@@ -34,6 +34,8 @@
 #include <event.h>
 #include <fcntl.h>
 #include <pwd.h>
+#include <sha1.h>
+#include <sha2.h>
 #include <signal.h>
 #include <syslog.h>
 #include <unistd.h>
@@ -41,6 +43,7 @@
 
 #include "got_opentemp.h"
 #include "got_reference.h"
+#include "got_object.h"
 
 #include "gotwebd.h"
 #include "log.h"
@@ -53,6 +56,7 @@ void	 gotwebd_configure_done(struct gotwebd *);
 void	 gotwebd_sighdlr(int sig, short event, void *arg);
 void	 gotwebd_shutdown(void);
 void	 gotwebd_dispatch_server(int, short, void *);
+void	 gotwebd_dispatch_fcgi(int, short, void *);
 void	 gotwebd_dispatch_gotweb(int, short, void *);
 
 struct gotwebd	*gotwebd_env;
@@ -119,16 +123,7 @@ int
 main_compose_sockets(struct gotwebd *env, uint32_t type, int fd,
     const void *data, uint16_t len)
 {
-	size_t i;
-	int ret = 0;
-
-	for (i = 0; i < env->nserver; ++i) {
-		ret = send_imsg(&env->iev_server[i], type, fd, data, len);
-		if (ret)
-			break;
-	}
-
-	return ret;
+	return send_imsg(env->iev_sockets, type, fd, data, len);
 }
 
 int
@@ -138,7 +133,7 @@ main_compose_gotweb(struct gotwebd *env, uint32_t type, int fd,
 	size_t i;
 	int ret = 0;
 
-	for (i = 0; i < env->nserver; ++i) {
+	for (i = 0; i < env->prefork; i++) {
 		ret = send_imsg(&env->iev_gotweb[i], type, fd, data, len);
 		if (ret)
 			break;
@@ -204,6 +199,56 @@ gotwebd_dispatch_server(int fd, short event, void *arg)
 	}
 }
 
+void
+gotwebd_dispatch_fcgi(int fd, short event, void *arg)
+{
+	struct imsgev		*iev = arg;
+	struct imsgbuf		*ibuf;
+	struct imsg		 imsg;
+	struct gotwebd		*env = gotwebd_env;
+	ssize_t			 n;
+	int			 shut = 0;
+
+	ibuf = &iev->ibuf;
+
+	if (event & EV_READ) {
+		if ((n = imsgbuf_read(ibuf)) == -1)
+			fatal("imsgbuf_read error");
+		if (n == 0)	/* Connection closed */
+			shut = 1;
+	}
+	if (event & EV_WRITE) {
+		if (imsgbuf_write(ibuf) == -1)
+			fatal("imsgbuf_write");
+	}
+
+	for (;;) {
+		if ((n = imsg_get(ibuf, &imsg)) == -1)
+			fatal("imsg_get");
+		if (n == 0)	/* No more messages. */
+			break;
+
+		switch (imsg.hdr.type) {
+		case GOTWEBD_IMSG_CFG_DONE:
+			gotwebd_configure_done(env);
+			break;
+		default:
+			fatalx("%s: unknown imsg type %d", __func__,
+			    imsg.hdr.type);
+		}
+
+		imsg_free(&imsg);
+	}
+
+	if (!shut)
+		imsg_event_add(iev);
+	else {
+		/* This pipe is dead.  Remove its event handler */
+		event_del(&iev->ev);
+		event_loopexit(NULL);
+	}
+}
+
 void
 gotwebd_dispatch_gotweb(int fd, short event, void *arg)
 {
@@ -283,7 +328,7 @@ spawn_process(struct gotwebd *env, const char *argv0, struct imsgev *iev,
     enum gotwebd_proc_type proc_type, const char *username,
     void (*handler)(int, short, void *))
 {
-	const char	*argv[9];
+	const char	*argv[10];
 	int		 argc = 0;
 	int		 p[2];
 	pid_t		 pid;
@@ -311,9 +356,17 @@ spawn_process(struct gotwebd *env, const char *argv0, struct imsgev *iev,
 	close(p[1]);
 
 	argv[argc++] = argv0;
-	if (proc_type == GOTWEBD_PROC_SERVER) {
+	if (proc_type == GOTWEBD_PROC_SOCKETS) {
+		char *s;
+
 		argv[argc++] = "-S";
 		argv[argc++] = username;
+		if (asprintf(&s, "-S%d", env->prefork) == -1)
+			fatal("asprintf");
+		argv[argc++] = s;
+	} else if (proc_type == GOTWEBD_PROC_FCGI) {
+		argv[argc++] = "-F";
+		argv[argc++] = username;
 	} else if (proc_type == GOTWEBD_PROC_GOTWEB) {
 		argv[argc++] = "-G";
 		argv[argc++] = username;
@@ -364,7 +417,7 @@ main(int argc, char **argv)
 	const char		*www_username = GOTWEBD_WWW_USER;
 	gid_t			 gotwebd_groups[NGROUPS_MAX];
 	gid_t			 www_gid;
-	const char		*argv0;
+	const char		*argv0, *errstr;
 
 	if ((argv0 = argv[0]) == NULL)
 		argv0 = "gotwebd";
@@ -377,7 +430,7 @@ main(int argc, char **argv)
 		fatal("%s: calloc", __func__);
 	config_init(env);
 
-	while ((ch = getopt(argc, argv, "D:dG:f:nS:vW:")) != -1) {
+	while ((ch = getopt(argc, argv, "D:dG:f:F:nS:vW:")) != -1) {
 		switch (ch) {
 		case 'D':
 			if (cmdline_symset(optarg) < 0)
@@ -394,12 +447,20 @@ main(int argc, char **argv)
 		case 'f':
 			conffile = optarg;
 			break;
+		case 'F':
+			proc_type = GOTWEBD_PROC_FCGI;
+			gotwebd_username = optarg;
+			break;
 		case 'n':
 			no_action = 1;
 			break;
 		case 'S':
-			proc_type = GOTWEBD_PROC_SERVER;
-			gotwebd_username = optarg;
+			proc_type = GOTWEBD_PROC_SOCKETS;
+			i = strtonum(optarg, 1, INT_MAX, &errstr);
+			if (errstr)
+				gotwebd_username = optarg;
+			else
+				env->prefork = i;
 			break;
 		case 'v':
 			if (env->gotwebd_verbose < 3)
@@ -432,6 +493,13 @@ main(int argc, char **argv)
 			www_username = env->www_user;
 	}
 
+	if (proc_type == GOTWEBD_PROC_SOCKETS) {
+		env->worker_load = calloc(env->prefork,
+		    sizeof(env->worker_load[0]));
+		if (env->worker_load == NULL)
+			fatal("calloc");
+	}
+
 	pw = getpwnam(www_username);
 	if (pw == NULL)
 		fatalx("unknown user %s", www_username);
@@ -452,9 +520,9 @@ main(int argc, char **argv)
 	log_setverbose(env->gotwebd_verbose);
 
 	switch (proc_type) {
-	case GOTWEBD_PROC_SERVER:
-		setproctitle("server");
-		log_procinit("server");
+	case GOTWEBD_PROC_SOCKETS:
+		setproctitle("sockets");
+		log_procinit("sockets");
 
 		if (chroot(env->httpd_chroot) == -1)
 			fatal("chroot %s", env->httpd_chroot);
@@ -468,6 +536,22 @@ main(int argc, char **argv)
 
 		sockets(env, GOTWEBD_SOCK_FILENO);
 		return 1;
+	case GOTWEBD_PROC_FCGI:
+		setproctitle("fcgi");
+		log_procinit("fcgi");
+
+		if (chroot(env->httpd_chroot) == -1)
+			fatal("chroot %s", env->httpd_chroot);
+		if (chdir("/") == -1)
+			fatal("chdir /");
+
+		if (setgroups(gotwebd_ngroups, gotwebd_groups) == -1 ||
+		    setresgid(pw->pw_gid, pw->pw_gid, pw->pw_gid) == -1 ||
+		    setresuid(pw->pw_uid, pw->pw_uid, pw->pw_uid) == -1)
+			fatal("failed to drop privileges");
+
+		gotwebd_fcgi(env, GOTWEBD_SOCK_FILENO);
+		return 1;
 	case GOTWEBD_PROC_GOTWEB:
 		setproctitle("gotweb");
 		log_procinit("gotweb");
@@ -488,22 +572,32 @@ main(int argc, char **argv)
 
 	evb = event_init();
 
-	env->nserver = env->prefork_gotwebd;
-	env->iev_server = calloc(env->nserver, sizeof(*env->iev_server));
-	if (env->iev_server == NULL)
+	env->iev_sockets = calloc(1, sizeof(*env->iev_sockets));
+	if (env->iev_sockets == NULL)
 		fatal("calloc");
-	env->iev_gotweb = calloc(env->nserver, sizeof(*env->iev_gotweb));
+
+	env->iev_fcgi = calloc(1, sizeof(*env->iev_fcgi));
+	if (env->iev_fcgi == NULL)
+		fatal("calloc");
+
+	env->iev_gotweb = calloc(env->prefork, sizeof(*env->iev_gotweb));
 	if (env->iev_gotweb == NULL)
 		fatal("calloc");
 
-	for (i = 0; i < env->nserver; ++i) {
-		spawn_process(env, argv0, &env->iev_server[i],
-		    GOTWEBD_PROC_SERVER, gotwebd_username,
-		    gotwebd_dispatch_server);
+	spawn_process(env, argv0, env->iev_sockets,
+	    GOTWEBD_PROC_SOCKETS, gotwebd_username,
+	    gotwebd_dispatch_server);
+
+	spawn_process(env, argv0, env->iev_fcgi,
+	    GOTWEBD_PROC_FCGI, gotwebd_username,
+	    gotwebd_dispatch_fcgi);
+
+	for (i = 0; i < env->prefork; ++i) {
 		spawn_process(env, argv0, &env->iev_gotweb[i],
 		    GOTWEBD_PROC_GOTWEB, gotwebd_username,
 		    gotwebd_dispatch_gotweb);
 	}
+
 	if (chdir("/") == -1)
 		fatal("chdir /");
 
@@ -558,21 +652,31 @@ main(int argc, char **argv)
 static void
 connect_children(struct gotwebd *env)
 {
-	struct imsgev *iev1, *iev2;
+	struct imsgev *iev_gotweb;
 	int pipe[2];
 	int i;
 
-	for (i = 0; i < env->nserver; i++) {
-		iev1 = &env->iev_server[i];
-		iev2 = &env->iev_gotweb[i];
+	if (socketpair(AF_UNIX, SOCK_STREAM, PF_UNSPEC, pipe) == -1)
+		fatal("socketpair");
+
+	if (main_compose_sockets(env, GOTWEBD_IMSG_CTL_PIPE, pipe[0], NULL, 0))
+		fatal("main_compose_sockets");
+
+	if (send_imsg(env->iev_fcgi, GOTWEBD_IMSG_CTL_PIPE, pipe[1], NULL, 0))
+		fatal("send_imsg");
+
+	for (i = 0; i < env->prefork; i++) {
+		iev_gotweb = &env->iev_gotweb[i];
 
 		if (socketpair(AF_UNIX, SOCK_STREAM, PF_UNSPEC, pipe) == -1)
 			fatal("socketpair");
 
-		if (send_imsg(iev1, GOTWEBD_IMSG_CTL_PIPE, pipe[0], NULL, 0))
+		if (main_compose_sockets(env, GOTWEBD_IMSG_CTL_PIPE,
+		    pipe[0], NULL, 0))
 			fatal("send_imsg");
 
-		if (send_imsg(iev2, GOTWEBD_IMSG_CTL_PIPE, pipe[1], NULL, 0))
+		if (send_imsg(iev_gotweb, GOTWEBD_IMSG_CTL_PIPE,
+		    pipe[1], NULL, 0))
 			fatal("send_imsg");
 	}
 }
@@ -584,14 +688,13 @@ gotwebd_configure(struct gotwebd *env, uid_t uid, gid_t gid)
 	struct socket *sock;
 
 	/* gotweb need to reload its config. */
-	env->servers_pending = env->prefork_gotwebd;
-	env->gotweb_pending = env->prefork_gotwebd;
+	env->gotweb_pending = env->prefork;
 
 	/* send our gotweb servers */
 	TAILQ_FOREACH(srv, &env->servers, entry) {
 		if (main_compose_sockets(env, GOTWEBD_IMSG_CFG_SRV,
 		    -1, srv, sizeof(*srv)) == -1)
-			fatal("main_compose_sockets GOTWEBD_IMSG_CFG_SRV");
+			fatal("send_imsg GOTWEBD_IMSG_CFG_SRV");
 		if (main_compose_gotweb(env, GOTWEBD_IMSG_CFG_SRV,
 		    -1, srv, sizeof(*srv)) == -1)
 			fatal("main_compose_gotweb GOTWEBD_IMSG_CFG_SRV");
@@ -610,8 +713,9 @@ gotwebd_configure(struct gotwebd *env, uid_t uid, gid_t gid)
 	/* Connect servers and gotwebs. */
 	connect_children(env);
 
-	if (main_compose_sockets(env, GOTWEBD_IMSG_CFG_DONE, -1, NULL, 0) == -1)
-		fatal("main_compose_sockets GOTWEBD_IMSG_CFG_DONE");
+	if (main_compose_sockets(env, GOTWEBD_IMSG_CFG_DONE, -1,
+	    NULL, 0) == -1)
+		fatal("send_imsg GOTWEBD_IMSG_CFG_DONE");
 
 	return (0);
 }
@@ -619,13 +723,9 @@ gotwebd_configure(struct gotwebd *env, uid_t uid, gid_t gid)
 void
 gotwebd_configure_done(struct gotwebd *env)
 {
-	if (env->servers_pending > 0) {
-		env->servers_pending--;
-		if (env->servers_pending == 0 &&
-		    main_compose_sockets(env, GOTWEBD_IMSG_CTL_START,
-		        -1, NULL, 0) == -1)
-			fatal("main_compose_sockets GOTWEBD_IMSG_CTL_START");
-	}
+	if (main_compose_sockets(env, GOTWEBD_IMSG_CTL_START,
+	    -1, NULL, 0) == -1)
+		fatal("send_imsg GOTWEBD_IMSG_CTL_START");
 
 	if (env->gotweb_pending > 0) {
 		env->gotweb_pending--;
@@ -643,15 +743,19 @@ gotwebd_shutdown(void)
 	pid_t		 pid;
 	int		 i, status;
 
-	for (i = 0; i < env->nserver; ++i) {
-		event_del(&env->iev_server[i].ev);
-		imsgbuf_clear(&env->iev_server[i].ibuf);
-		close(env->iev_server[i].ibuf.fd);
-		env->iev_server[i].ibuf.fd = -1;
-	}
-	free(env->iev_server);
+	event_del(&env->iev_sockets->ev);
+	imsgbuf_clear(&env->iev_sockets->ibuf);
+	close(env->iev_sockets->ibuf.fd);
+	env->iev_sockets->ibuf.fd = -1;
+	free(env->iev_sockets);
 
-	for (i = 0; i < env->nserver; ++i) {
+	event_del(&env->iev_fcgi->ev);
+	imsgbuf_clear(&env->iev_fcgi->ibuf);
+	close(env->iev_fcgi->ibuf.fd);
+	env->iev_fcgi->ibuf.fd = -1;
+	free(env->iev_fcgi);
+
+	for (i = 0; i < env->prefork; ++i) {
 		event_del(&env->iev_gotweb[i].ev);
 		imsgbuf_clear(&env->iev_gotweb[i].ibuf);
 		close(env->iev_gotweb[i].ibuf.fd);
@@ -679,6 +783,13 @@ gotwebd_shutdown(void)
 		TAILQ_REMOVE(&gotwebd_env->addresses, h, entry);
 		free(h);
 	}
+	while (!TAILQ_EMPTY(&gotwebd_env->sockets)) {
+		struct socket *sock;
+
+		sock = TAILQ_FIRST(&gotwebd_env->sockets);
+		TAILQ_REMOVE(&gotwebd_env->sockets, sock, entry);
+		free(sock);
+	}
 	while (!TAILQ_EMPTY(&gotwebd_env->servers)) {
 		struct server *srv;
 
