@@ -123,7 +123,8 @@ struct got_reflist_head;
 
 enum gotwebd_proc_type {
 	GOTWEBD_PROC_PARENT,
-	GOTWEBD_PROC_SERVER,
+	GOTWEBD_PROC_SOCKETS,
+	GOTWEBD_PROC_FCGI,
 	GOTWEBD_PROC_GOTWEB,
 };
 
@@ -134,6 +135,9 @@ enum imsg_type {
 	GOTWEBD_IMSG_CFG_DONE,
 	GOTWEBD_IMSG_CTL_PIPE,
 	GOTWEBD_IMSG_CTL_START,
+	GOTWEBD_IMSG_FCGI_PARSE_PARAMS,
+	GOTWEBD_IMSG_FCGI_PARAMS,
+	GOTWEBD_IMSG_REQ_ABORT,
 	GOTWEBD_IMSG_REQ_PROCESS,
 };
 
@@ -208,7 +212,7 @@ struct transport {
 	struct got_reflist_head	 refs;
 	struct got_repository	*repo;
 	struct repo_dir		*repo_dir;
-	struct querystring	*qs;
+	const struct querystring *qs;
 	char			*more_id;
 	char			*tags_more_id;
 	unsigned int		 repos_total;
@@ -239,6 +243,48 @@ enum socket_priv_fds {
 	PRIV_FDS__MAX,
 };
 
+struct gotwebd_fcgi_record {
+	uint32_t			 request_id;
+	uint8_t				 record[FCGI_RECORD_SIZE];
+	size_t				 record_len;
+};
+
+enum query_actions {
+	NO_ACTION = 0,
+	BLAME,
+	BLOB,
+	BLOBRAW,
+	BRIEFS,
+	COMMITS,
+	DIFF,
+	ERR,
+	INDEX,
+	PATCH,
+	SUMMARY,
+	TAG,
+	TAGS,
+	TREE,
+	RSS,
+};
+
+struct querystring {
+	enum query_actions action;
+	char		 commit[GOT_OBJECT_ID_HEX_MAXLEN];
+	char		 file[NAME_MAX];
+	char		 folder[PATH_MAX];
+	char		 headref[MAX_DOCUMENT_URI];
+	int		 index_page;
+	char		 path[PATH_MAX];
+};
+
+struct gotwebd_fcgi_params {
+	uint32_t			 request_id;
+	struct querystring		 qs;
+	char				 document_uri[MAX_DOCUMENT_URI];
+	char				 server_name[MAX_SERVER_NAME];
+	int				 https;
+};
+
 struct template;
 struct request {
 	TAILQ_ENTRY(request)		entry;
@@ -252,22 +298,19 @@ struct request {
 	uint16_t			 id;
 	int				 fd;
 	int				 priv_fd[PRIV_FDS__MAX];
-	struct event			*resp_event;
 	int				 sock_id;
 	uint32_t			 request_id;
+	int				 worker_idx;
 
 	uint8_t				 *buf;
-	size_t				 buf_pos;
 	size_t				 buf_len;
 
 	uint8_t				 *outbuf;
 
-	char				 querystring[MAX_QUERYSTRING];
-	char				 document_uri[MAX_DOCUMENT_URI];
-	char				 server_name[MAX_SERVER_NAME];
-	int				 https;
+	struct gotwebd_fcgi_params	 fcgi_params;
+	int				 nparams;
+	int				 nparams_parsed;
 
-	uint8_t				 request_started;
 	int				 client_status;
 };
 TAILQ_HEAD(requestlist, request);
@@ -325,6 +368,9 @@ TAILQ_HEAD(serverlist, server);
 
 enum client_action {
 	CLIENT_CONNECT,
+	CLIENT_FCGI_BEGIN,
+	CLIENT_FCGI_PARAMS,
+	CLIENT_FCGI_STDIN,
 	CLIENT_REQUEST,
 	CLIENT_DISCONNECT,
 };
@@ -366,15 +412,13 @@ struct gotwebd {
 	int		 gotwebd_verbose;
 
 	struct imsgev	*iev_parent;
-	struct imsgev	*iev_server;
+	struct imsgev	*iev_sockets;
+	struct imsgev	*iev_fcgi;
 	struct imsgev	*iev_gotweb;
-	size_t		 nserver;
 
-	uint16_t	 prefork_gotwebd;
-	int		 servers_pending;
+	uint16_t	 prefork;
 	int		 gotweb_pending;
-
-	int		 server_cnt;
+	int		 *worker_load;
 
 	char		 httpd_chroot[PATH_MAX];
 };
@@ -393,16 +437,6 @@ struct gotweb_url {
 	const char	*path;
 };
 
-struct querystring {
-	uint8_t		 action;
-	char		*commit;
-	char		*file;
-	char		*folder;
-	char		*headref;
-	int		 index_page;
-	char		*path;
-};
-
 struct querystring_keys {
 	const char	*name;
 	int		 element;
@@ -423,23 +457,6 @@ enum querystring_elements {
 	PATH,
 };
 
-enum query_actions {
-	BLAME,
-	BLOB,
-	BLOBRAW,
-	BRIEFS,
-	COMMITS,
-	DIFF,
-	ERR,
-	INDEX,
-	PATCH,
-	SUMMARY,
-	TAG,
-	TAGS,
-	TREE,
-	RSS,
-};
-
 extern struct gotwebd	*gotwebd_env;
 
 typedef int (*got_render_blame_line_cb)(struct template *, const char *,
@@ -501,11 +518,11 @@ int parse_config(const char *, struct gotwebd *);
 int cmdline_symset(char *);
 
 /* fcgi.c */
-void fcgi_request(int, short, void *);
-void fcgi_timeout(int, short, void *);
+void fcgi_init_querystring(struct querystring *);
 void fcgi_cleanup_request(struct request *);
 void fcgi_create_end_record(struct request *);
 int fcgi_write(void *, const void *, size_t);
+void gotwebd_fcgi(struct gotwebd *, int);
 
 /* got_operations.c */
 const struct got_error *got_gotweb_closefile(FILE *);
