@@ -248,6 +248,14 @@ static const struct got_error got_errors[] = {
 	{ GOT_ERR_BAD_KEYWORD, "invalid commit keyword" },
 	{ GOT_ERR_UNKNOWN_CAPA, "unknown capability" },
 	{ GOT_ERR_REF_DUP_ENTRY, "duplicate reference entry" },
+	{ GOT_ERR_DIFF_NOCHANGES, "no changes match the requested diff" },
+	{ GOT_ERR_USER, "no such user" },
+	{ GOT_ERR_USER_EXISTS, "user already exists" },
+	{ GOT_ERR_GROUP, "no such group" },
+	{ GOT_ERR_GROUP_EXISTS, "group already exists" },
+	{ GOT_ERR_AUTHORIZED_KEY, "no authorized key found" },
+	{ GOT_ERR_CONNECTION_LIMIT, "connection limit exceeded" },
+	{ GOT_ERR_ON_SERVER_SIDE, "see server-side logs for error details" },
 };
 
 static struct got_custom_error {
@@ -267,8 +275,6 @@ got_error(int code)
 {
 	size_t i;
 
-	if (code == GOT_ERR_PRIVSEP_NO_FD) abort();
-
 	for (i = 0; i < nitems(got_errors); i++) {
 		if (code == got_errors[i].code)
 			return &got_errors[i];
